{
   Currently support the following chipsets:

 S3           Tested OK!  (ViRGE 3D, TRIO 64v+)
 TRIDENT      Tested OK!
 TSENG        Tested OK!  (TSeng ET6000)
 CIRRUS       Tested OK!  (Cirrus 5465)
 ATI          Tested OK!
 ARK          !Not Tested
 COMPAQ       !Not Tested
 OAK          !Not Tested
 SIERRA       !Not Tested
 ACER         !Not Tested
 MACH64       Same as ATI
 VIDEO7       !Not Tested
 SiS          !Not Tested
 Paradise     !Detection not implemented
}

procedure outpb(p,d:dword);
  begin
    Port[p]:=d;
  end;

procedure outpw(p,d:dword);
  begin
    PortW[p]:=d;
  end;

procedure outpd(p,d:dword);
  begin
    PortD[p]:=d;
  end;

function inpb(p:dword):dword;
  begin
    inpb:=Port[p];
  end;

function inpw(p:dword):dword;
  begin
    inpw:=PortW[p];
  end;

function inpd(p:dword):dword;
  begin
    inpd:=PortD[p];
  end;

procedure WriteByte(p,index,data:dword);
  begin
    outpb(p,index);
    outpb(p+1,data);
  end;

function  ReadByte(p,index:dword):dword;
  begin
    outpb(p,index);
    ReadByte:=inpb(p+1);
  end;

procedure WriteWord(p,index,data:dword);
  begin
    outpw(p,index);
    outpw(p+1,data);
  end;

function  ReadWord(p,index:dword):dword;
  begin
    outpw(p,index);
    ReadWord:=inpw(p+1);
  end;

procedure S3_enable;
  begin
    { Unlock S3 extended registers }
    WriteByte($3D4, $38, $48);
    WriteByte($3D4, $39, $0A5);
    { write to Linear Address Window Control Register }
    WriteByte($3D4,$58,  ReadByte($3D4,$58) OR $10);
  end;

procedure Trident_enable;
  begin
    { Enable New Mode Registers   }
    ReadByte($3C4,$0B);
    { enable extended registers   }
    WriteByte($3C4,$0E,  ReadByte($3C4,$0E) OR $80);
    { write to Configurable Linear Addressing Register }
    WriteByte($3D4,$21,  ReadByte($3D4,$21) OR $20);
  end;

procedure Tseng_enable;
  begin
    { write to Video System Configuration 1 Register }
    WriteByte($3D4,$36,  ReadByte($3D4,$36) OR $10);
  end;

procedure Cirrus_enable;
  begin
    { write to CPU Base Address Control Register }
    WriteByte($3CE,$0D,  ReadByte($3CE,$0D) OR $10);
  end;

procedure ATI_enable;
  begin
    { write to ATI Register }
    WriteByte($1CE,$0AC, ReadByte($1CE,$0AC) OR 01);
  end;

procedure ACER_enable;
  begin
    WriteByte($3CE,$17,  ReadByte($3CE,$17) OR $80);
  end;

procedure ARK_enable;
  begin
    { Enable extended registers }
    WriteByte($3C4,$1D,  ReadByte($3C4,$1D) OR 1);
    { linear frame buffer size 4 MB }
    WriteByte($3C4,$12,  ReadByte($3C4,$12) OR $03);
    { Enable access to whole video memory and enable LFB }
    WriteByte($3C4,$10,  ReadByte($3C4,$10) OR $13);
//    WriteByte($3C4,$10,  ReadByte($3C4,$10) OR $10);
  end;

procedure COMPAQ_enable;
  begin
    outpw($23C7,  inpw($23C7) OR 4);  { ?? }
  end;

procedure OAK_enable;
  begin
    { write to Video Memory Mapping Register }
    WriteByte($3DE,$05,  ReadByte($3DE,$05) OR 01);
  end;

{?? how i can detect this ??}
{3D4 29h?? }
procedure Paradise_enable;
  begin
    { unlock Paradise Extended Registers }
    WriteByte($3C4,$06,  $48);
    { write to Video Memory Mapping Register }
    WriteByte($3C4,$14,  ReadByte($3C4,$14) OR $30); {?? set both bits ??}
  end;

procedure Sierra_enable;
  begin
    WriteByte($3CE,$05,  ReadByte($3CE,$05) OR $80);
  end;

procedure Sis_enable;
  begin
    { write to Video Memory Mapping Register }
    WriteWord($3C4,$20,  ReadWord($3C4,$20) OR $6000); { set bits 13-14 }
  end;

function SiS_detect:boolean;
  begin
     SiS_detect:=false;
     WriteByte($3C4, $05, 00);    { Lock extended registers }
     if ReadByte($3C4, $05)<>$21 then exit;    { must return $21, not SiS }
     WriteByte($3C4, $05, $86);   { UnLock extended registers }
     if ReadByte($3C4, $05)<>$0A1 then exit;   { must return $A1, not SiS }
  end;

function  MACH64_detect:boolean;
  var regs:TRmRegs;
      mach64_floating,mach64_iobase:word;
      scratch_reg:longint;
      old:dword;
  function get_mach64_port(io_sel, mm_sel: longint):longint;
    begin
      if (mach64_floating<>0) then
           get_mach64_port:= (mm_sel shl 2) + mach64_iobase
      else get_mach64_port:= (io_sel shl 10) + mach64_iobase;
    end;
  begin
     { this detection routine comletely ripped from Mach 64 driver of FreeBE }
     { project }
     ClearRmRegs(regs);
     { try to detect present of Mach64 bios }
     regs.ax:=$0A012;
     regs.cx:=0;
     RealModeInt($10,regs);
     { should return zero }
     MACH64_detect:= (regs.ah=0);
     if regs.ah<>0 then exit;  { ok, we don't have Mach64 }

     mach64_floating := regs.cx;
     mach64_iobase := regs.dx;

     scratch_reg := get_mach64_port($11, $21);
     old := inpd(scratch_reg);

     outpd(scratch_reg, $55555555);
     if (inpd(scratch_reg) <> $55555555) then begin
       outpd(scratch_reg, old);
       MACH64_detect:=false;
       exit;
     end;

     outpd(scratch_reg, $0AAAAAAAA);
     if (inpd(scratch_reg) <> $0AAAAAAAA) then begin
       outpd(scratch_reg, old);
       MACH64_detect:=false;
       exit;
     end;
     outpd(scratch_reg, old);
     { we have confirmed the presence of a mach64 }
     MACH64_detect:=true;
  end;

procedure Mach64_enable;
  var regs:TRmRegs;
  begin
     ClearRmRegs(regs);
     { enable LFB by calling Mach64 BIOS }
     regs.ax:=$0A005;
     regs.cx:=1;
     RealModeInt($10,regs);
  end;

function VIDEO7_detect:boolean;
  var regs:TRmRegs;
      old:dword;
  begin
     { this detection routine comletely ripped from FreeBE project }

   old := ReadByte($3C4, 6);
   WriteByte($3C4, 6, $0EA);   { enable BIOS extentions }

   ClearRmRegs(regs);
   regs.ax := $6F00;                            { installation check }
   regs.bx := $0000;
   RealModeInt($10, regs);
   if ((regs.bx <> $5637) AND (regs.bx <> $4850)) then begin
      WriteByte($3C4, 6, old);
      VIDEO7_detect:=false;
   end else VIDEO7_detect:=true;
  end;

procedure VIDEO7_enable;
  begin
    { write to Miscellaneous Control 2 Register }
    WriteByte($3C4,$0C8,  ReadByte($3C4,$0C8) OR $40);
  end;

