unit VFlat;
Interface

uses DPMI,crt,strings;
CONST
  vferrOK               =  0;
  vferrVFNotAvaible     = -1;
  vferrModeNotSupported = -2;
  vferrErrorSettingMode = -3;

function  vfSetMode(mode:word):integer;
function  vfActive:boolean;
function  vfGetBaseAddress:dword;
function  vfGetPhysAddress:dword;
procedure vfCloseMode;
function  vfForceSoftware:boolean;

var
    vfGlobalError:integer;

implementation

type call_proc=procedure;
const
   PCI_ID_STRING = $20494350;  {' ICP'}
   PCI_INT       = $1A;

   { known vendor IDs }
   PCI_VENDOR_ID_S3      = $5333;
   PCI_VENDOR_ID_TRIDENT = $1023;
   PCI_VENDOR_ID_TSENG   = $100C;
   PCI_VENDOR_ID_CIRRUS  = $1013;
   PCI_VENDOR_ID_ATI     = $1002;
   PCI_VENDOR_ID_ARK     = $0EDD8;
   PCI_VENDOR_ID_COMPAQ  = $0E11;
   PCI_VENDOR_ID_OAK     = $104E;
   PCI_VENDOR_ID_SIERRA  = $10A8;
   PCI_VENDOR_ID_ACER    = $1025;  { is this right code? }

   PCI_VENDOR_ID_NCR     = $1000;  { !!Dont have initialization code }
   PCI_VENDOR_ID_WEITEK  = $100e;  { !!Dont have initialization code }

   VFLAT_BASE            = $0F0000000; // software base address
   _software             = 1;
   _hardware             = 0;

var VFAvaible :boolean;
    PCIAvaible:boolean;
    VFModeInitiated:boolean;
    PCI_ver   :dword;  // pci version in BCD format
    PCISurface :dword;
    PCI_Vendor:word;
    vfInitType:integer;  // wich way mode was initiated
    vfLFB_seg :dword;    // linear address of mapped framebuffer
    TotalVideoMemory:dword;  // amount of video memory

{$I vftypes.pas}
{$L vfasm.obj}
{$I chips.pas}
Function  InitVFlat:integer;external;
Procedure DoneVFlat;external;
procedure InitPageTable(banksize:dword);external;
procedure DonePageTable;external;
procedure InstallVFHandler(banksize:dword);external;
procedure RemoveVFHandler;external;

function  FindDeviceInitProc(var CP:call_proc):boolean;
 var InitProc:pointer;
 begin
    InitProc:=nil;
    case PCI_vendor of
         PCI_VENDOR_ID_S3      : InitProc:=@S3_enable;
         PCI_VENDOR_ID_TRIDENT : InitProc:=@Trident_enable;
         PCI_VENDOR_ID_TSENG   : InitProc:=@TSeng_enable;
         PCI_VENDOR_ID_CIRRUS  : InitProc:=@Cirrus_enable;
         PCI_VENDOR_ID_ATI     : InitProc:=@ATI_enable;
         // here is goes non - tested vide cards
         PCI_VENDOR_ID_ARK     : InitProc:=@ARK_enable;
         PCI_VENDOR_ID_COMPAQ  : InitProc:=@COMPAQ_enable;
         PCI_VENDOR_ID_OAK     : InitProc:=@OAK_enable;
         PCI_VENDOR_ID_SIERRA  : InitProc:=@SIERRA_enable;
         PCI_VENDOR_ID_ACER    : InitProc:=@ACER_enable;

         PCI_VENDOR_ID_NCR,
         PCI_VENDOR_ID_WEITEK  : begin
                                   // we don't have initialization code for those chipsets
                                   TextMode(co80);
                                   writeln('unsupported chipset detected!! please report to <mit@pc2o205a.pnpi.spb.ru>');
                                   writeln('now press any key to continue...');
                                   readkey;
                                 end;
    end;
    { perform additional checking for known devices }
    if Mach64_detect then InitProc:=@Mach64_enable;  {?? is this a same as ATI??}
    if Video7_detect then InitProc:=@Video7_enable;
    if SiS_detect    then InitProc:=@SiS_enable;

    if ((PCI_vendor = PCI_VENDOR_ID_ARK)    OR
        (PCI_vendor = PCI_VENDOR_ID_COMPAQ) OR
        (PCI_vendor = PCI_VENDOR_ID_OAK)    OR
        (PCI_vendor = PCI_VENDOR_ID_SIERRA) OR
        (PCI_vendor = PCI_VENDOR_ID_ACER)   OR
         Video7_detect                      OR
         SiS_detect) then                             begin
         TextMode(co80);
         writeln('untested chipset detected!! please report to <mit@pc2o205a.pnpi.spb.ru>');
         writeln('graphics initialization can fail and you will not be able to see anything');
         writeln('now press any key to continue...');
         readkey;
    end;

    CP:=call_proc(initproc); { forget about this warning }
    FindDeviceInitProc:= (initproc<>nil);
 end;

function  vfForceSoftware:boolean;
  begin
     // think that PCI was not detected to activate software emulation
     // but if software not avaible then leave hardware.
     if VFAvaible then begin
        PCIAvaible:=false;
        PCISurface:=0;
     end;
     vfForceSoftware:=VFAvaible;
  end;

function  vfGetPhysAddress:dword;
  begin
    vfGetPhysAddress:=PCISurface;
  end;

function vfGetBaseAddress:dword;
  begin
    vfGetBaseAddress:=vfLFB_seg;
  end;

Function VESA_GetModeInfo(Mode:word; var InfoBlock:TVbeModeInfo):boolean;
 var Regs:TRmRegs;
 begin
    {this must be done for some VESA implementations}
    FillChar(pointer(buf_32)^,SizeOf(TVbeModeInfo),$00000000);
    FillChar(InfoBlock,SizeOf(InfoBlock),0);

    ClearRmRegs(Regs);
    Regs.ES:=buf_16;  {use internal 8k buffer}
    Regs.AX:=$4F01;
    Regs.CX:=mode and $3FFF;  {clear all flags}
    RealModeInt($10,Regs);
    { if mode supported? }
    if (Regs.AX<>$004F) then
       VESA_GetModeInfo:=false
    else begin
       VESA_GetModeInfo:=true;
       Move(Pointer(buf_32)^,InfoBlock,SizeOf(TVbeModeInfo));
    end;
 end;

Function VESA_detect(var info:TVbeInfo):boolean;
  var Regs:TRmRegs;
  begin
     {force vbe to give VESA20 extended info}
     Info.VBESignature[1]:='V';
     Info.VBESignature[2]:='B';
     Info.VBESignature[3]:='E';
     Info.VBESignature[4]:='2';
     {this must be done for some VESA implementations}
     FillChar(pointer(buf_32)^,SizeOf(TVbeInfo),$00000000);
     {copy first 4 bytes that contain VBE2 string into temporary block}
     move(Info,pointer(buf_32)^,4);
     {check vesa compatibility}
     ClearRmRegs(Regs);
     Regs.ax:=$4F00;
     Regs.es:=buf_16;
     RealModeInt($10,Regs);
     move(pointer(buf_32)^,info,SizeOf(TVbeInfo));
     if (Regs.ax<>$004F) or ((info.VBESignature[1]<>'V') OR
                             (info.VBESignature[2]<>'E') OR
                             (info.VBESignature[3]<>'S') OR
                             (info.VBESignature[4]<>'A')) then begin
              VESA_detect:=false;
     end else VESA_detect:=true;
  end;

function vfSetMode(mode:word):integer;
  label _TrySoftware;
  var ModeInfo:TVbeModeInfo;
      regs:TRmRegs;
      lfb_init:call_proc;
      linear:boolean:=false;
  begin
     vfSetMode:=vferrOK;
     vfGlobalError:=0;
     // first close previous initiated mode
     if VFModeInitiated then vfCloseMode;
     // get information about this mode
     if not VESA_GetModeInfo(mode,ModeInfo) then begin
        vfSetMode:=vferrModeNotSupported;
        exit;
     end;
     // if linear mode supported by VBE?
     linear:= (((ModeInfo.ModeAttributes AND $80)<>0) AND
                (ModeInfo.PhysBasePtr<>0));
     // if PCI is avaible and surface is detected or linear framebuffer
     // supported by VBE then just init this mode
     if (PCIAvaible OR Linear) then begin
        if not linear then begin
            // if we only have PCI
            // search for supported video adapter
            if not FindDeviceInitProc(lfb_init) then
               // this is unsupported device. Try to use software emulation
               GoTo _TrySoftware;
        end;
        ClearRMRegs(regs);
        regs.ax:=$4F02;
        // set LFB flag if linear is avaible
        if linear then mode:=mode OR $4000;
        regs.bx:=mode;
        RealModeInt($10,regs);
        if (regs.AX<>$004F) then begin
     // store error code returned by VBE to allow debugging from main program
           vfGlobalError:=regs.ax;
           vfSetMode:=vferrErrorSettingMode;
           exit;
        end else
           VFModeInitiated:=true;
        // we using hardware linear framebuffer
        vfInitType:=_hardware;
        // map physical video memory to linear
        if linear then
           // use address provided by VBE to be sure... Set limit to
           // amount of video memory
           vfLFB_seg:=MapPhysicalToLinear(ModeInfo.PhysBasePtr,TotalVideoMemory)
        else begin
           // if LFB not supported by VBE then initiate it using
           // hardware-specific code
           lfb_init;
           // map memory provided by PCI. Set limit to 4Mb
           vfLFB_seg:=MapPhysicalToLinear(PCISurface,4048*1024);
        end;
        exit;
     end;
     // we don't have PCI video adapter and LFB doesn't supported by VBE
     // try to initiate software emulation
_TrySoftware:
     if (not VFAvaible) then begin
        // software emulation not avaible, nothing more to do here...
        vfSetMode:=vferrVFNotAvaible;
        exit;
     end;

     if (ModeInfo.WinGranularity<>64) AND (ModeInfo.WinGranularity<>4) then begin
        // VFlat currently supports only 64K banks
        vfSetMode:=vferrModeNotSupported;
        exit;
     end;

     ClearRMRegs(regs);
     regs.ax:=$4F02;
     regs.bx:=mode;  { LFB flag doesn't needed for software emulation }
     RealModeInt($10,regs);
     if (regs.AX<>$004F) then begin
        vfSetMode:=vferrErrorSettingMode;
        exit;
     end;
     // initiate paging table
     InitPageTable(ModeInfo.WinGranularity);
     InstallVFHandler(ModeInfo.WinGranularity);
     VFModeInitiated:=True;
     // software emulation
     vfInitType:=_software;
     // we already have linear address, so we don't need to map it.
     vfLFB_seg:=VFLAT_BASE;
  end;

procedure vfCloseMode;
  var regs:TRmRegs;
  begin
     // if mode wasn't initiated...
     if not VFModeInitiated then exit;
     VFModeInitiated:=false;
     // restore text mode
     ClearRmRegs(regs);
     regs.ax:=3;
     RealModeInt($10,regs);
     if vfInitType=_software then begin
         // remove software emulation handler
         RemoveVFHandler;
         DonePageTable;
     end else
     if vfInitType=_hardware then
         // free physical mapping
         FreePhysicalMap(vfLFB_seg);
  end;

// some PCI functions
{$I vfpci.pas}

function vfActive:boolean;
  begin
    // software emulation is active...
    vfActive:= (vfInitType=_software);
  end;

var OldExitProc:pointer;
procedure _VFExitProc;
   begin
      ExitProc:=OldExitProc;
      vfCloseMode;
      // if software vflat was initiated then deinit it...
      if VFAvaible then DoneVFlat;
   end;

var VBEinfo:TVbeInfo;
begin
  if not VESA_detect(VBEInfo) then begin
      // vesa was not detected... fatal
      PCIAvaible:=false; VFAvaible:=false;
      exit;
  end;
  // calculate amount of video memory in bytes
  TotalVideoMemory:=VBEInfo.TotalMemory*65536;

  VFModeInitiated:=false;
  OldExitProc:=ExitProc;
  ExitProc:=@_VFExitProc;   // append to exit proc chain
  PCIAvaible:=PCI_detect;
  // if PCI BIOS detected then try to find VGA compatible adapter with
  // linear framebuffer
  if PCIAvaible then PCISurface:=PCI_FindLFBSurface else PCISurface:=0;
  // this must never happen...but who knows...
  if (PCIAvaible AND (PCISurface=0)) then PCIAvaible:=false;
  // also try initiate virtual flat framebuffer
  VFAvaible:= (InitVFlat=0);
end.